@@ -14,10 +14,11 @@ import platform
 import subprocess
 
 import numpy as np
-from scipy.sparse import isspmatrix
+import scipy.sparse as sp
 from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
 from sklearn.exceptions import NotFittedError
 from sklearn.externals import six
+from sklearn.externals.joblib import Parallel, delayed
 from sklearn.utils.extmath import softmax
 from sklearn.utils.multiclass import check_classification_targets
 from sklearn.utils.validation import check_array, check_consistent_length, check_X_y, column_or_1d
@@ -71,20 +72,59 @@ def _get_paths():
     return def_exe, exe, temp
 
 
+_DEFAULT_EXE_PATH, _EXE_PATH, _TEMP_PATH = _get_paths()
+
+
+if not os.path.isdir(_TEMP_PATH):
+    os.makedirs(_TEMP_PATH)
+if not os.access(_TEMP_PATH, os.W_OK):
+    raise Exception("{0} is not writable directory. Please set "
+                    "config flag 'temp_location' to writable directory".format(_TEMP_PATH))
+
+
 def _is_executable_response(path):
+    temp_x_loc = os.path.join(_TEMP_PATH, 'temp.train.data.x')
+    temp_y_loc = os.path.join(_TEMP_PATH, 'temp.train.data.y')
+    np.savetxt(temp_x_loc, [[1, 0, 1, 0], [0, 1, 0, 1]], delimiter=' ', fmt="%s")
+    np.savetxt(temp_y_loc, [1, -1], delimiter=' ', fmt="%s")
+    _UUIDS.append('temp')
+    params = []
+    params.append("train_x_fn=%s" % temp_x_loc)
+    params.append("train_y_fn=%s" % temp_y_loc)
+    params.append("model_fn_prefix=%s" % os.path.join(_TEMP_PATH, "temp.model"))
+    params.append("reg_L2=%s" % 1)
+
     try:
-        subprocess.check_output((path, "train"))
+        subprocess.check_output((path, "train", ",".join(params)))
         return True
     except Exception:
         return False
 
 
+if _is_executable_response(_DEFAULT_EXE_PATH):
+    _EXE_PATH = _DEFAULT_EXE_PATH
+elif _is_executable_response(os.path.join(os.path.dirname(__file__), _DEFAULT_EXE_PATH)):
+    _EXE_PATH = os.path.join(os.path.dirname(__file__), _DEFAULT_EXE_PATH)
+elif not os.path.isfile(_EXE_PATH):
+    raise Exception("{0} is not executable file. Please set "
+                    "config flag 'exe_location' to RGF execution file.".format(_EXE_PATH))
+elif not os.access(_EXE_PATH, os.X_OK):
+    raise Exception("{0} cannot be accessed. Please set "
+                    "config flag 'exe_location' to RGF execution file.".format(_EXE_PATH))
+elif _is_executable_response(_EXE_PATH):
+    pass
+else:
+    raise Exception("{0} does not exist or {1} is not in the "
+                    "'PATH' variable.".format(_EXE_PATH, _DEFAULT_EXE_PATH))
+
+
 @atexit.register
 def _cleanup():
-    for uuid in _UUIDS:
-        model_glob = os.path.join(_TEMP_PATH, uuid + "*")
-        for fn in glob(model_glob):
-            os.remove(fn)
+    if _UUIDS is not None:
+        for uuid in _UUIDS:
+            model_glob = os.path.join(_TEMP_PATH, uuid + "*")
+            for fn in glob(model_glob):
+                os.remove(fn)
 
 
 def _sigmoid(x):
@@ -105,7 +145,9 @@ def _validate_params(max_leaf,
                      opt_interval,
                      learning_rate,
                      verbose,
-                     calc_prob="Sigmoid"):
+                     memory_policy,
+                     calc_prob="sigmoid",
+                     n_jobs=-1):
     if not isinstance(max_leaf, (numbers.Integral, np.integer)):
         raise ValueError("max_leaf must be an integer, got {0}.".format(type(max_leaf)))
     elif max_leaf <= 0:
@@ -179,19 +221,26 @@ def _validate_params(max_leaf,
     elif verbose < 0:
         raise ValueError("verbose must be no smaller than 0 but was %r." % verbose)
 
+    if not isinstance(memory_policy, six.string_types):
+        raise ValueError("memory_policy must be a string, got {0}.".format(type(memory_policy)))
+    elif memory_policy not in ("conservative", "generous"):
+        raise ValueError("memory_policy must be 'conservative' or 'generous' but was %r." % memory_policy)
+
     if not isinstance(calc_prob, six.string_types):
         raise ValueError("calc_prob must be a string, got {0}.".format(type(calc_prob)))
-    elif calc_prob not in ("Sigmoid", "Softmax"):
-        raise ValueError("calc_prob must be 'Sigmoid' or 'Softmax' but was %r." % calc_prob)
+    elif calc_prob not in ("sigmoid", "softmax"):
+        raise ValueError("calc_prob must be 'sigmoid' or 'softmax' but was %r." % calc_prob)
+
+    if not isinstance(n_jobs, (numbers.Integral, np.integer)):
+        raise ValueError("n_jobs must be an integer, got {0}.".format(type(n_jobs)))
 
 
 def _sparse_savetxt(filename, input_array):
-    try:  # For Python 2.x
-        from itertools import izip
-        zip_func = izip
-    except ImportError:  # For Python 3.x
-        zip_func = zip
-    input_array = input_array.tocsr().tocoo()
+    zip_func = six.moves.zip
+    if sp.isspmatrix_csr(input_array):
+        input_array = input_array.tocoo()
+    elif not sp.isspmatrix_coo(input_array):
+        input_array = input_array.tocsr().tocoo()
     n_row = input_array.shape[0]
     current_sample_row = 0
     line = []
@@ -209,6 +258,10 @@ def _sparse_savetxt(filename, input_array):
         fw.write('\n' * (n_row - i))
 
 
+def _fit_ovr_binary(binary_clf, X, y, sample_weight):
+    return binary_clf.fit(X, y, sample_weight)
+
+
 class _AtomicCounter(object):
     def __init__(self):
         self.value = 0
@@ -221,28 +274,6 @@ class _AtomicCounter(object):
 
 
 _COUNTER = _AtomicCounter()
-_DEFAULT_EXE_PATH, _EXE_PATH, _TEMP_PATH = _get_paths()
-
-if _is_executable_response(_DEFAULT_EXE_PATH):
-    _EXE_PATH = _DEFAULT_EXE_PATH
-elif not os.path.isfile(_EXE_PATH):
-    raise Exception("{0} is not executable file. Please set "
-                    "config flag 'exe_location' to RGF execution file.".format(_EXE_PATH))
-elif not os.access(_EXE_PATH, os.X_OK):
-    raise Exception("{0} cannot be accessed. Please set "
-                    "config flag 'exe_location' to RGF execution file.".format(_EXE_PATH))
-elif _is_executable_response(_EXE_PATH):
-    pass
-else:
-    raise Exception("{0} does not exist or {1} is not in the "
-                    "'PATH' variable.".format(_EXE_PATH, _DEFAULT_EXE_PATH))
-
-if not os.path.isdir(_TEMP_PATH):
-    os.makedirs(_TEMP_PATH)
-if not os.access(_TEMP_PATH, os.W_OK):
-    raise Exception("{0} is not writable directory. Please set "
-                    "config flag 'temp_location' to writable directory".format(_TEMP_PATH))
-
 
 
 class RGFClassifier(BaseEstimator, ClassifierMixin):
@@ -250,7 +281,7 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
     A Regularized Greedy Forest [1] classifier.
 
     Tuning parameters detailed instruction:
-        http://tongzhang-ml.org/software/rgf/rgf1.2-guide.pdf
+        https://github.com/fukatani/rgf_python/blob/master/include/rgf/rgf1.2-guide.pdf
 
     Parameters
     ----------
@@ -269,6 +300,9 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
 
     loss : string ("LS" or "Expo" or "Log"), optional (default="Log")
         Loss function.
+        LS: Square loss.
+        Expo: Exponential loss.
+        Log: Logistic loss.
 
     reg_depth : float, optional (default=1.0)
         Must be no smaller than 1.0.
@@ -312,12 +346,26 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
     learning_rate : float, optional (default=0.5)
         Step size of Newton updates used in coordinate descent to optimize weights.
 
+    calc_prob : string ("sigmoid" or "softmax"), optional (default="sigmoid")
+        Method of probability calculation.
+
+    n_jobs : integer, optional (default=-1)
+        The number of jobs to use for the computation.
+        If 1 is given, no parallel computing code is used at all.
+        If -1 all CPUs are used.
+        For n_jobs = -2, all CPUs but one are used.
+        For n_jobs below -1, (n_cpus + 1 + n_jobs) are used.
+
+    memory_policy : string ("conservative" or "generous"), optional (default="generous")
+        Memory using policy.
+        Generous: it runs faster using more memory by keeping the sorted orders
+        of the features on memory for reuse.
+        Conservative: it uses less memory at the expense of longer runtime. Try only when
+        with default value it uses too much memory.
+
     verbose : int, optional (default=0)
         Controls the verbosity of the tree building process.
 
-    calc_prob : string ("Sigmoid" or "Softmax"), optional (default="Sigmoid")
-        Method of probability calculation.
-
     Attributes:
     -----------
     estimators_ : list of binary classifiers
@@ -354,8 +402,10 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
                  n_tree_search=1,
                  opt_interval=100,
                  learning_rate=0.5,
-                 verbose=0,
-                 calc_prob='Sigmoid'):
+                 calc_prob="sigmoid",
+                 n_jobs=-1,
+                 memory_policy="generous",
+                 verbose=0):
         self.max_leaf = max_leaf
         self.test_interval = test_interval
         self.algorithm = algorithm
@@ -369,9 +419,10 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
         self.n_tree_search = n_tree_search
         self.opt_interval = opt_interval
         self.learning_rate = learning_rate
-        self.verbose = verbose
         self.calc_prob = calc_prob
-        self.fitted_ = False
+        self.n_jobs = n_jobs
+        self.memory_policy = memory_policy
+        self.verbose = verbose
 
     def fit(self, X, y, sample_weight=None):
         """
@@ -415,6 +466,7 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
 
         X, y = check_X_y(X, y, accept_sparse=True)
         n_samples, self.n_features_ = X.shape
+
         if sample_weight is None:
             sample_weight = np.ones(n_samples, dtype=np.float32)
         else:
@@ -427,48 +479,46 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
         self.classes_ = sorted(np.unique(y))
         self.n_classes_ = len(self.classes_)
         self._classes_map = {}
+
+        params = dict(max_leaf=self.max_leaf,
+                      test_interval=self.test_interval,
+                      algorithm=self.algorithm,
+                      loss=self.loss,
+                      reg_depth=self.reg_depth,
+                      l2=self.l2,
+                      sl2=self.sl2_,
+                      normalize=self.normalize,
+                      min_samples_leaf=self.min_samples_leaf_,
+                      n_iter=self.n_iter_,
+                      n_tree_search=self.n_tree_search,
+                      opt_interval=self.opt_interval,
+                      learning_rate=self.learning_rate,
+                      memory_policy=self.memory_policy,
+                      verbose=self.verbose)
         if self.n_classes_ == 2:
             self._classes_map[0] = self.classes_[0]
             self._classes_map[1] = self.classes_[1]
             self.estimators_ = [None]
             y = (y == self.classes_[0]).astype(int)
-            self.estimators_[0] = _RGFBinaryClassifier(max_leaf=self.max_leaf,
-                                                       test_interval=self.test_interval,
-                                                       algorithm=self.algorithm,
-                                                       loss=self.loss,
-                                                       reg_depth=self.reg_depth,
-                                                       l2=self.l2,
-                                                       sl2=self.sl2_,
-                                                       normalize=self.normalize,
-                                                       min_samples_leaf=self.min_samples_leaf_,
-                                                       n_iter=self.n_iter_,
-                                                       n_tree_search=self.n_tree_search,
-                                                       opt_interval=self.opt_interval,
-                                                       learning_rate=self.learning_rate,
-                                                       verbose=self.verbose)
+            self.estimators_[0] = _RGFBinaryClassifier(**params)
             self.estimators_[0].fit(X, y, sample_weight)
         elif self.n_classes_ > 2:
+            if sp.isspmatrix_dok(X):
+                X = X.tocsr().tocoo()  # Fix to avoid scipy 7699 issue
             self.estimators_ = [None] * self.n_classes_
+            ovr_list = [None] * self.n_classes_
             for i, cls_num in enumerate(self.classes_):
                 self._classes_map[i] = cls_num
-                y_one_or_rest = (y == cls_num).astype(int)
-                self.estimators_[i] = _RGFBinaryClassifier(max_leaf=self.max_leaf,
-                                                           test_interval=self.test_interval,
-                                                           algorithm=self.algorithm,
-                                                           loss=self.loss,
-                                                           reg_depth=self.reg_depth,
-                                                           l2=self.l2,
-                                                           sl2=self.sl2_,
-                                                           normalize=self.normalize,
-                                                           min_samples_leaf=self.min_samples_leaf_,
-                                                           n_iter=self.n_iter_,
-                                                           n_tree_search=self.n_tree_search,
-                                                           opt_interval=self.opt_interval,
-                                                           learning_rate=self.learning_rate,
-                                                           verbose=self.verbose)
-                self.estimators_[i].fit(X, y_one_or_rest, sample_weight)
+                ovr_list[i] = (y == cls_num).astype(int)
+                self.estimators_[i] = _RGFBinaryClassifier(**params)
+            self.estimators_ = Parallel(n_jobs=self.n_jobs)(delayed(_fit_ovr_binary)(self.estimators_[i],
+                                                                                     X,
+                                                                                     ovr_list[i],
+                                                                                     sample_weight)
+                                                            for i in range(self.n_classes_))
         else:
             raise ValueError("Classifier can't predict when only one class is present.")
+
         self.fitted_ = True
         return self
 
@@ -489,7 +539,7 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
             The class probabilities of the input samples.
             The order of the classes corresponds to that in the attribute classes_.
         """
-        if not self.fitted_:
+        if not hasattr(self, 'fitted_') or not self.fitted_:
             raise NotFittedError("Estimator not fitted, "
                                  "call `fit` before exploiting the model.")
         X = check_array(X, accept_sparse=True)
@@ -500,25 +550,25 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
                              "input n_features is %s "
                              % (self.n_features_, n_features))
         if self.n_classes_ == 2:
-            proba = self.estimators_[0].predict_proba(X)
-            proba = _sigmoid(proba)
-            proba = np.c_[proba, 1 - proba]
+            y = self.estimators_[0].predict_proba(X)
+            y = _sigmoid(y)
+            y = np.c_[y, 1 - y]
         else:
-            proba = np.zeros((X.shape[0], self.n_classes_))
+            y = np.zeros((X.shape[0], self.n_classes_))
             for i, clf in enumerate(self.estimators_):
                 class_proba = clf.predict_proba(X)
-                proba[:, i] = class_proba
+                y[:, i] = class_proba
 
-            # In honest I don't understand which is better
+            # In honest, I don't understand which is better
             # softmax or normalized sigmoid for calc probability.
-            if self.calc_prob == "Sigmoid":
-                proba = _sigmoid(proba)
-                normalizer = proba.sum(axis=1)[:, np.newaxis]
+            if self.calc_prob == "sigmoid":
+                y = _sigmoid(y)
+                normalizer = np.sum(y, axis=1)[:, np.newaxis]
                 normalizer[normalizer == 0.0] = 1.0
-                proba /= normalizer
+                y /= normalizer
             else:
-                proba = softmax(proba)
-        return proba
+                y = softmax(y)
+        return y
 
     def predict(self, X):
         """
@@ -536,9 +586,9 @@ class RGFClassifier(BaseEstimator, ClassifierMixin):
         y : array of shape = [n_samples]
             The predicted classes.
         """
-        proba = self.predict_proba(X)
-        y_pred = np.argmax(proba, axis=1)
-        return np.asarray(list(self._classes_map.values()))[np.searchsorted(list(self._classes_map.keys()), y_pred)]
+        y = self.predict_proba(X)
+        y = np.argmax(y, axis=1)
+        return np.asarray(list(self._classes_map.values()))[np.searchsorted(list(self._classes_map.keys()), y)]
 
 
 class _RGFBinaryClassifier(BaseEstimator, ClassifierMixin):
@@ -561,6 +611,7 @@ class _RGFBinaryClassifier(BaseEstimator, ClassifierMixin):
                  n_tree_search=1,
                  opt_interval=100,
                  learning_rate=0.5,
+                 memory_policy="generous",
                  verbose=0):
         self.max_leaf = max_leaf
         self.test_interval = test_interval
@@ -575,16 +626,16 @@ class _RGFBinaryClassifier(BaseEstimator, ClassifierMixin):
         self.n_tree_search = n_tree_search
         self.opt_interval = opt_interval
         self.learning_rate = learning_rate
+        self.memory_policy = memory_policy
         self.verbose = verbose
         self._file_prefix = str(uuid4()) + str(_COUNTER.increment())
         _UUIDS.append(self._file_prefix)
-        self.fitted_ = False
 
-    def fit(self, X, y, sample_weight=None):
+    def fit(self, X, y, sample_weight):
         train_x_loc = os.path.join(_TEMP_PATH, self._file_prefix + ".train.data.x")
         train_y_loc = os.path.join(_TEMP_PATH, self._file_prefix + ".train.data.y")
         train_weight_loc = os.path.join(_TEMP_PATH, self._file_prefix + ".train.data.weight")
-        if isspmatrix(X):
+        if sp.isspmatrix(X):
             _sparse_savetxt(train_x_loc, X)
         else:
             np.savetxt(train_x_loc, X, delimiter=' ', fmt="%s")
@@ -598,6 +649,8 @@ class _RGFBinaryClassifier(BaseEstimator, ClassifierMixin):
         params = []
         if self.verbose > 0:
             params.append("Verbose")
+        if self.verbose > 5:
+            params.append("Verbose_opt")  # Add some info on weight optimization
         if self.normalize:
             params.append("NormalizeTarget")
         params.append("train_x_fn=%s" % train_x_loc)
@@ -614,6 +667,7 @@ class _RGFBinaryClassifier(BaseEstimator, ClassifierMixin):
         params.append("num_tree_search=%s" % self.n_tree_search)
         params.append("opt_interval=%s" % self.opt_interval)
         params.append("opt_stepsize=%s" % self.learning_rate)
+        params.append("memory_policy=%s" % self.memory_policy.title())
         params.append("model_fn_prefix=%s" % os.path.join(_TEMP_PATH, self._file_prefix + ".model"))
         params.append("train_w_fn=%s" % train_weight_loc)
 
@@ -633,12 +687,12 @@ class _RGFBinaryClassifier(BaseEstimator, ClassifierMixin):
         return self
 
     def predict_proba(self, X):
-        if not self.fitted_:
+        if not hasattr(self, 'fitted_') or not self.fitted_:
             raise NotFittedError("Estimator not fitted, "
                                  "call `fit` before exploiting the model.")
 
         test_x_loc = os.path.join(_TEMP_PATH, self._file_prefix + ".test.data.x")
-        if isspmatrix(X):
+        if sp.isspmatrix(X):
             _sparse_savetxt(test_x_loc, X)
         else:
             np.savetxt(test_x_loc, X, delimiter=' ', fmt="%s")
@@ -669,8 +723,7 @@ class _RGFBinaryClassifier(BaseEstimator, ClassifierMixin):
             for k in output:
                 print(k)
 
-        y_pred = np.loadtxt(pred_loc)
-        return y_pred
+        return np.loadtxt(pred_loc)
 
 
 class RGFRegressor(BaseEstimator, RegressorMixin):
@@ -678,7 +731,7 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
     A Regularized Greedy Forest [1] regressor.
 
     Tuning parameters detailed instruction:
-        http://tongzhang-ml.org/software/rgf/rgf1.2-guide.pdf
+        https://github.com/fukatani/rgf_python/blob/master/include/rgf/rgf1.2-guide.pdf
 
     Parameters
     ----------
@@ -740,6 +793,13 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
     learning_rate : float, optional (default=0.5)
         Step size of Newton updates used in coordinate descent to optimize weights.
 
+    memory_policy : string ("conservative" or "generous"), optional (default="generous")
+        Memory using policy.
+        Generous: it runs faster using more memory by keeping the sorted orders
+        of the features on memory for reuse.
+        Conservative: it uses less memory at the expense of longer runtime. Try only when
+        with default value it uses too much memory.
+
     verbose : int, optional (default=0)
         Controls the verbosity of the tree building process.
 
@@ -770,6 +830,7 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
                  n_tree_search=1,
                  opt_interval=100,
                  learning_rate=0.5,
+                 memory_policy="generous",
                  verbose=0):
         self.max_leaf = max_leaf
         self.test_interval = test_interval
@@ -784,10 +845,10 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
         self.n_tree_search = n_tree_search
         self.opt_interval = opt_interval
         self.learning_rate = learning_rate
+        self.memory_policy = memory_policy
         self.verbose = verbose
         self._file_prefix = str(uuid4()) + str(_COUNTER.increment())
         _UUIDS.append(self._file_prefix)
-        self.fitted_ = False
 
     def fit(self, X, y, sample_weight=None):
         """
@@ -842,7 +903,7 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
         train_x_loc = os.path.join(_TEMP_PATH, self._file_prefix + ".train.data.x")
         train_y_loc = os.path.join(_TEMP_PATH, self._file_prefix + ".train.data.y")
         train_weight_loc = os.path.join(_TEMP_PATH, self._file_prefix + ".train.data.weight")
-        if isspmatrix(X):
+        if sp.isspmatrix(X):
             _sparse_savetxt(train_x_loc, X)
         else:
             np.savetxt(train_x_loc, X, delimiter=' ', fmt="%s")
@@ -853,6 +914,8 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
         params = []
         if self.verbose > 0:
             params.append("Verbose")
+        if self.verbose > 5:
+            params.append("Verbose_opt")  # Add some info on weight optimization
         if self.normalize:
             params.append("NormalizeTarget")
         params.append("train_x_fn=%s" % train_x_loc)
@@ -869,6 +932,7 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
         params.append("num_tree_search=%s" % self.n_tree_search)
         params.append("opt_interval=%s" % self.opt_interval)
         params.append("opt_stepsize=%s" % self.learning_rate)
+        params.append("memory_policy=%s" % self.memory_policy.title())
         params.append("model_fn_prefix=%s" % os.path.join(_TEMP_PATH, self._file_prefix + ".model"))
         params.append("train_w_fn=%s" % train_weight_loc)
 
@@ -903,7 +967,7 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
         y : array of shape = [n_samples]
             The predicted values.
         """
-        if not self.fitted_:
+        if not hasattr(self, 'fitted_') or not self.fitted_:
             raise NotFittedError("Estimator not fitted, "
                                  "call `fit` before exploiting the model.")
 
@@ -916,7 +980,7 @@ class RGFRegressor(BaseEstimator, RegressorMixin):
                              % (self.n_features_, n_features))
 
         test_x_loc = os.path.join(_TEMP_PATH, self._file_prefix + ".test.data.x")
-        if isspmatrix(X):
+        if sp.isspmatrix(X):
             _sparse_savetxt(test_x_loc, X)
         else:
             np.savetxt(test_x_loc, X, delimiter=' ', fmt="%s")
